
import cx.ath.butiki.coffeetalk.*;
import java.util.*;
import java.net.*;
import java.io.*;

public class CoffeeTalk extends Peer implements Runnable, Observer
{

    private List dialogues;
    private String peer_name;
    private PeerManager peer_man;
    private DialogueServer server;

    public CoffeeTalk(String pname)
    {
	peer_name = pname;
	InetAddress b_addr;
	try{
	    b_addr = InetAddress.getByName("235.0.0.5");
	}
	catch(UnknownHostException e){
	    System.err.println("CoffeeTalk: Couldn't lookup broadcast address");
	    b_addr = null;
	    System.exit(255);
	}
	
	peer_man = new PeerManager(this, b_addr);
	dialogues = new ArrayList();
	server = new DialogueServer(this);
	server.start();
	peer_man.addObserver(this);
    }

    public String getPeerName()
    {
	return peer_name;
    }

    public String[] getPeerCapabilities()
    {
	String[] caps = { "im", "chat", "diagnostics" };
	return caps;
    }

    /*
     */
    public boolean handleDialogue(Dialogue d)
    {
	// add to dialogue list
	//System.out.println("Dialogue started: " + d.getOthersName());
	dialogues.add(d);
	Thread t = new MessageListener(d);
	t.start();
	return true;
    }

    public void update(Observable o, Object arg)
    {
	// List was changed
	System.out.println("Peer network changed status");
    }
    
    // Session thread
    public void run()
    {
	boolean keep_alive = true;
	BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
	String command;
	
	// wait for dialogue
        try{
	    while(keep_alive){
		System.out.print(":> ");
		command = in.readLine();
		if(command.equalsIgnoreCase("talk-to")){
		    System.out.print("** Name: ");
		    String peername = in.readLine();
		    Peer p = peer_man.getPeer(peername);
		    if(p != null){
			// go do the deed; create a dialogue
			handleDialogue(new Dialogue(this, p));
		    }
		    else{
			System.err.println("Error: No peer by that name found");
		    }
		}
		else if(command.equalsIgnoreCase("list")){
		    Peer[] peers = peer_man.getPeers();
		    System.out.println("** Peers: ");
		    if(peers != null){
			for(int i = 0; i < peers.length; i++){
			    System.out.print("\t");
			    System.out.print(peers[i].getPeerName());
			    System.out.print("(");
			    System.out.print(peers[i].getPeerAddress());
			    System.out.println(")");
			}
		    }
		    else{
			System.out.println("** No peers in network");
		    }
		}
		else if(command.equalsIgnoreCase("whoami")){
		    System.out.println("** You are:");
		    System.out.println(this.getPeerName());
		}
		else if(command.equalsIgnoreCase("msg")){
		    System.out.print("** Message: ");
		    String message = in.readLine();
		    // broadcast to all connected dialogue peers
		    if(!dialogues.isEmpty()){
			for(Iterator i = dialogues.iterator(); i.hasNext(); ){
			    Dialogue d = (Dialogue) i.next();
			    d.writeMessage(message);
			}
		    }
		    else
			System.out.println("** Not broadcasted. No one to talk to.");
		}
		else if(command.equalsIgnoreCase("quit")){
		    System.out.println("** Stopping...");
		    keep_alive = false;
		    peer_man.StopPeerManager();
		    server.terminateMe();
		}
		else{
		    System.out.println("ERR: Not a valid command");
		}
	    }
        }
	catch(IOException e){
	    System.err.println("IOException");
	    e.printStackTrace(System.err);
	}
	
	System.out.println("** Thread exiting...");
    }

    public static void main(String[] args)
    {
	String name = "barako!";
	// First arg is peer name?
	if(args.length >= 1){
	    name = args[0];
	}

	System.out.println("Coffee Talk peer-to-peer system");
	System.out.println("(C) UA&P Team Coffee");
	System.out.println("All rights reserved");
	System.out.println("");
	System.out.println("CoffeeTalk reference");
	System.out.prinltn("client implementation v0.1");
	System.out.println("--------------------------");

	CoffeeTalk cf_peer = new CoffeeTalk(name);

	// Creates the running session thread
	Thread session = new Thread(cf_peer);
	System.out.print("Starting peer session thread...");
	session.start();
	System.out.println("Peer started.");
    }

    public InetAddress getPeerAddress()
    {
	try{
	    return InetAddress.getLocalHost();
	}
	catch(UnknownHostException e){
	    // impossible; we're trying to get the local address
	    return null;
	}
    }
    
    /** Query this peer's resources.
     */
    public Resource[] getPeerResources() {
	return null;
    }
    
    /**
     * Monitors for dialogue connections
     */
    private class DialogueServer extends Thread
    {
	private ServerSocket listener;
	private Peer parent;
	
	public DialogueServer(Peer who)
	{
	    parent = who;
	    try{
		listener = new ServerSocket(Dialogue.DIALOGUE_PORT);
	    }
	    catch(IOException e){
		System.err.println("DialogueServer: Exception binding to dialogue socket");
	    }
	}
	
	public void run()
	{
	    // wait
	    Socket client = null;
	    try{
		listener.setSoTimeout(100);
	    }
	    catch(SocketException ignore){ }
	    while(still_alive){
		try{
		    client = listener.accept();
		    if(client != null)
			parent.handleDialogue(new Dialogue(parent, client));
		    client = null;
		}
		catch(Exception e){
		    // System.err.println("DialogueServer: Exception thrown while listening");
		}
	    }
	    System.out.println("Dialogue server exiting");
	}
	
	public void terminateMe()
	{
	    still_alive = false;
	    try{
		listener.close();
	    }
	    catch(Exception ignore){ }
	}
	
	private boolean still_alive = true;
    }
    
    private class MessageListener extends Thread
    {
	public MessageListener(Dialogue d)
	{
	    listen_to = d;
	}
	
	public void run()
	{
	    try{
		while(true){
		    String msg = listen_to.getMessage();
		    System.out.println(listen_to.getOthersName() + ": " + msg);
		}
	    }
	    catch(Exception stop){
		System.out.println(listen_to.getOthersName() + ": Stopped listening to dialogue!");
	    }
	}
	
	private Dialogue listen_to;
    }
}
